@@ -37,6 +37,9 @@
 -export([handle_call/3, handle_cast/2, handle_info/2, terminate/2, 
 	 code_change/3, init_starter/4, get_loaded/1]).
 
+%% logger callback
+-export([format_log/1, format_log/2]).
+
 %% Test exports, only to be used from the test suites
 -export([test_change_apps/2]).
 
@@ -155,7 +158,7 @@
 %% Env         = [{Key, Value}]
 %%-----------------------------------------------------------------
 
--record(appl, {name, appl_data, descr, id, vsn, restart_type, inc_apps, apps}).
+-record(appl, {name, appl_data, descr, id, vsn, restart_type, inc_apps, opt_apps, apps}).
 
 %%-----------------------------------------------------------------
 %% Func: start/1
@@ -268,13 +271,12 @@ which_applications(Timeout) ->
     gen_server:call(?AC, which_applications, Timeout).
 
 loaded_applications() ->
-    ets:filter(ac_tab,
-	       fun([{{loaded, AppName}, #appl{descr = Descr, vsn = Vsn}}]) ->
-		       {true, {AppName, Descr, Vsn}};
-		  (_) ->
-		       false
-	       end,
-	       []).
+    ets:select(ac_tab,
+               [{
+                  {{loaded, '$1'}, #appl{descr = '$2', vsn = '$3', _ = '_'}},
+                  [],
+                  [{{'$1', '$2', '$3'}}]
+                }]).
 
 %% Returns some debug info
 info() ->
@@ -371,6 +373,8 @@ get_key(AppName, Key) ->
 		    {ok, (Appl#appl.appl_data)#appl_data.regs};
 		included_applications ->
 		    {ok, Appl#appl.inc_apps};
+                optional_applications ->
+                    {ok, Appl#appl.opt_apps};
 		applications ->
 		    {ok, Appl#appl.apps};
 		env ->
@@ -402,6 +406,7 @@ get_all_key(AppName) ->
 		  {maxT, (Appl#appl.appl_data)#appl_data.maxT},
 		  {registered, (Appl#appl.appl_data)#appl_data.regs},
 		  {included_applications, Appl#appl.inc_apps},
+                  {optional_applications, Appl#appl.opt_apps},
 		  {applications, Appl#appl.apps},
 		  {env, get_all_env(AppName)},
 		  {mod, (Appl#appl.appl_data)#appl_data.mod},
@@ -522,6 +527,12 @@ init(Init, Kernel) ->
 		lists:flatten(io_lib:format("error in config file "
 					    "~tp (~w): ~ts",
 					    [File, Line, Str])),
+	    Init ! {ack, self(), {error, to_string(ReasonStr)}};
+        {error, {file_descriptor, FDString, Line, Str}} ->
+	    ReasonStr =
+		lists:flatten(io_lib:format("error in config read from file descriptor "
+					    "~tp (~w): ~ts",
+					    [FDString, Line, Str])),
 	    Init ! {ack, self(), {error, to_string(ReasonStr)}}
     end.
 
@@ -813,13 +824,11 @@ handle_call({stop_application, AppName}, _From, S) ->
     end;
 
 handle_call({change_application_data, Applications, Config}, _From, S) ->
-    OldAppls = ets:filter(ac_tab,
-			  fun([{{loaded, _AppName}, Appl}]) ->
-				  {true, Appl};
-			     (_) ->
-				  false
-			  end,
-			  []),
+    OldAppls = ets:foldl(fun({{loaded, _AppName}, Appl}, Acc) ->
+                                 [Appl|Acc];
+                            (_, Acc) ->
+                                 Acc
+                         end, [], ac_tab),
     case catch do_change_apps(Applications, Config, OldAppls) of
 	{error, _} = Error ->
 	    {reply, Error, S};
@@ -925,7 +934,7 @@ handle_application_started(AppName, Res, S) ->
     #state{starting = Starting, running = Running, started = Started, 
 	   start_req = Start_req} = S,
     Start_reqN = reply_to_requester(AppName, Start_req, Res),
-    {_AppName, RestartType, _Type, _From} = lists:keyfind(AppName, 1, Starting),
+    {AppName, RestartType, _Type, _From} = lists:keyfind(AppName, 1, Starting),
     case Res of
 	{ok, Id} ->
 	    case AppName of
@@ -951,8 +960,8 @@ handle_application_started(AppName, Res, S) ->
 			true ->
 			    #state{running = StopRunning, started = StopStarted} = NewS,
 			    case lists:keyfind(AppName, 1, StopRunning) of
-				{_AppName, Id} ->
-				    {_AppName2, Type} =
+				{AppName, Id} ->
+				    {AppName, Type} =
 					lists:keyfind(AppName, 1, StopStarted),
 				    stop_appl(AppName, Id, Type),
 				    NStopRunning = keydelete(AppName, 1, StopRunning),
@@ -1269,15 +1278,15 @@ do_load_application(Application, S) ->
 
 %% Recursively load the application and its included apps.
 %load(S, {ApplData, ApplEnv, IncApps, Descr, Vsn, Apps}) ->
-load(S, {ApplData, ApplEnv, IncApps, Descr, Id, Vsn, Apps}) ->
+load(S, {ApplData, ApplEnv, IncApps, OptApps, Descr, Id, Vsn, Apps}) ->
     Name = ApplData#appl_data.name,
     ConfEnv = get_env_i(Name, S),
     NewEnv = merge_app_env(ApplEnv, ConfEnv),
     CmdLineEnv = get_cmd_env(Name),
     NewEnv2 = merge_app_env(NewEnv, CmdLineEnv),
     add_env(Name, NewEnv2),
-    Appl = #appl{name = Name, descr = Descr, id = Id, vsn = Vsn, 
-		 appl_data = ApplData, inc_apps = IncApps, apps = Apps},
+    Appl = #appl{name = Name, descr = Descr, id = Id, vsn = Vsn, apps = Apps,
+		 appl_data = ApplData, inc_apps = IncApps, opt_apps = OptApps},
     ets:insert(ac_tab, {{loaded, Name}, Appl}),
     NewS =
 	foldl(fun(App, S1) ->
@@ -1314,13 +1323,13 @@ check_start_cond(AppName, RestartType, Started, Running) ->
 		    {error, {already_started, AppName}};
 		false ->
 		    foreach(
-		      fun(AppName2) ->
-			      case lists:keymember(AppName2, 1, Started) of
-				  true -> ok;
-				  false ->
-				      throw({error, {not_started, AppName2}})
-			      end
-		      end, Appl#appl.apps),
+			fun(AppName2) ->
+			    case lists:keymember(AppName2, 1, Started) orelse
+				     lists:member(AppName2, Appl#appl.opt_apps) of
+				true -> ok;
+				false -> throw({error, {not_started, AppName2}})
+			    end
+		    end, Appl#appl.apps),
 		    {ok, Appl}
 	    end;
 	false ->
@@ -1377,14 +1386,13 @@ start_appl(Appl, S, Type) ->
 	    %% Name = ApplData#appl_data.name,
 	    Running = S#state.running,
 	    foreach(
-	      fun(AppName) ->
-		      case lists:keymember(AppName, 1, Running) of
-			  true ->
-			      ok;
-			  false ->
-			      throw({info, {not_running, AppName}})
-		      end
-	      end, Appl#appl.apps),
+		fun(AppName) ->
+		    case lists:keymember(AppName, 1, Running) orelse
+			     lists:member(AppName, Appl#appl.opt_apps) of
+			true -> ok;
+			false -> throw({info, {not_running, AppName}})
+		    end
+		end, Appl#appl.apps),
 	    case application_master:start_link(ApplData, Type) of
 		{ok, _Pid} = Ok ->
 		    Ok;
@@ -1518,9 +1526,10 @@ make_appl_i({application, Name, Opts}) when is_atom(Name), is_list(Opts) ->
     MaxP = get_opt(maxP, Opts, infinity),
     MaxT = get_opt(maxT, Opts, infinity),
     IncApps = get_opt(included_applications, Opts, []),
+    OptApps = get_opt(optional_applications, Opts, []),
     {#appl_data{name = Name, regs = Regs, mod = Mod, phases = Phases,
-		mods = Mods, inc_apps = IncApps, maxP = MaxP, maxT = MaxT},
-     Env, IncApps, Descr, Id, Vsn, Apps};
+		mods = Mods, maxP = MaxP, maxT = MaxT},
+     Env, IncApps, OptApps, Descr, Id, Vsn, Apps};
 make_appl_i({application, Name, Opts}) when is_list(Opts) ->
     throw({error,{invalid_name,Name}});
 make_appl_i({application, _Name, Opts}) ->
@@ -1572,7 +1581,7 @@ is_loaded_app(AppName, [{application, AppName, App} | _]) ->
 is_loaded_app(AppName, [_ | T]) -> is_loaded_app(AppName, T);
 is_loaded_app(_AppName, []) -> false.
 
-do_change_appl({ok, {ApplData, Env, IncApps, Descr, Id, Vsn, Apps}},
+do_change_appl({ok, {ApplData, Env, IncApps, OptApps, Descr, Id, Vsn, Apps}},
 	       OldAppl, Config) ->
     AppName = OldAppl#appl.name,
 
@@ -1593,6 +1602,7 @@ do_change_appl({ok, {ApplData, Env, IncApps, Descr, Id, Vsn, Apps}},
 		 id=Id,
 		 vsn=Vsn,
 		 inc_apps=IncApps,
+                 opt_apps=OptApps,
 		 apps=Apps};
 do_change_appl({error, _R} = Error, _Appl, _ConfData) ->
     throw(Error).
@@ -1797,48 +1807,109 @@ do_config_diff([{Env, Value} | AppEnvNow], AppEnvBefore, {Changed, New}) ->
     end.
 
 
+conf_param_to_conf({config, FileName}) ->
+    BFName = filename:basename(FileName,".config"),
+    FName = filename:join(filename:dirname(FileName),
+                          BFName ++ ".config"),
+    case load_file(FName) of
+        {ok, NewEnv} ->
+            %% OTP-4867 sys.config may now contain names of other
+            %% .config files as well as configuration parameters.
+            %% Therefore read and merge contents.
+            if
+                BFName =:= "sys" ->
+                    DName = filename:dirname(FName),
+                    {ok, SysEnv, Errors} =
+                        check_conf_sys(NewEnv, [], [], DName),
+                    %% Report first error, if any, and terminate
+                    %% (backwards compatible behaviour)
+                    case Errors of
+                        [] ->
+                            SysEnv;
+                        [{error, {SysFName, Line, Str}}|_] ->
+                            throw({error, {SysFName, Line, Str}})
+                    end;
+                true ->
+                    NewEnv
+            end;
+        {error, {Line, _Mod, Str}} ->
+            throw({error, {FName, Line, Str}})
+    end;
+conf_param_to_conf({configfd, FileDescStrP}) ->
+    FileDescStr = unicode:characters_to_nfc_list(FileDescStrP),
+    IsDigit = fun(C) -> lists:member(C, "0123456789") end,
+    {FdString, FileDescType} = lists:splitwith(IsDigit, FileDescStr),
+    FileDesc =
+        try list_to_integer(FdString)
+        catch
+            error:badarg ->
+                throw({error, {file_descriptor,
+                               FileDescStr,
+                               invalid_file_desc,
+                               "The given file descriptor has incorrect format. "
+                               "The format should be \"FileDescId[.FileType]\". "
+                               "Examples: 3 or 3.config"}})
+        end,
+    case lists:member(FileDescType, [".config", ""])  of
+        false ->
+            throw({error, {file_descriptor,
+                           FileDescStr,
+                           invalid_file_desc,
+                           io_lib:format("Cannot parse file descriptor of type: ~ts",
+                                         [FileDescType])}});
+        true -> ok
+    end,
+    case load_file_descriptor(FileDesc) of
+        {ok, NewEnv} ->
+            %% Load config parameters from included config files
+            BootScript = init:script_name(),
+            DName = filename:dirname(BootScript),
+            {ok, SysEnv, Errors} =
+                check_conf_sys(NewEnv, [], [], DName),
+            case Errors of
+                [] ->
+                    SysEnv;
+                [{error, {SysFName, Line, Str}}|_] ->
+                    throw({error, {SysFName, Line, Str}})
+            end;
+        {error, {Line, _Mod, Str}} ->
+            throw({error, {file_descriptor,
+                           FileDescStr,
+                           Line,
+                           Str}})
+    end.
+
+config_param_to_list({Type, [First | _] = ConfigVals})
+  when
+      is_list(First),
+      Type =:= config orelse Type =:= configfd ->
+    [{Type, Val} || Val <- ConfigVals];
+config_param_to_list({config, Val}) ->
+    [{config, Val}];
+config_param_to_list({configfd, Val}) ->
+    [{configfd, Val}];
+config_param_to_list(_) ->
+    [].
+
+
+
 %%-----------------------------------------------------------------
 %% Read the .config files.
 %%-----------------------------------------------------------------
 check_conf() ->
-    case init:get_argument(config) of
-	{ok, Files} ->
-	    {ok, lists:foldl(
-		   fun([File], Env) ->
-			   BFName = filename:basename(File,".config"),
-			   FName = filename:join(filename:dirname(File),
-						 BFName ++ ".config"),
-			   case load_file(FName) of
-			       {ok, NewEnv} ->
-				   %% OTP-4867
-				   %% sys.config may now contain names of
-				   %% other .config files as well as
-				   %% configuration parameters.
-				   %% Therefore read and merge contents.
-				   if
-				       BFName =:= "sys" ->
-					   DName = filename:dirname(FName),
-					   {ok, SysEnv, Errors} =
-					       check_conf_sys(NewEnv, [], [], DName),
+    ConfigParameters =
+        lists:flatmap(
+          fun config_param_to_list/1,
+          init:get_arguments()),
+    MergedConf =
+        lists:foldl(fun (ConfigParameter, Env) ->
+                            NewEnv = conf_param_to_conf(ConfigParameter),
+                            merge_env(Env, NewEnv)
+                    end,
+                    [],
+                    ConfigParameters),
+    {ok, MergedConf}.
 
-					   %% Report first error, if any, and
-					   %% terminate
-					   %% (backwards compatible behaviour)
-					   case Errors of
-					       [] ->
-						   merge_env(Env, SysEnv);
-					       [{error, {SysFName, Line, Str}}|_] ->
-						   throw({error, {SysFName, Line, Str}})
-					   end;
-				       true ->
-					   merge_env(Env, NewEnv)
-				   end;
-			       {error, {Line, _Mod, Str}} ->
-				   throw({error, {FName, Line, Str}})
-			   end
-		   end, [], Files)};
-	_ -> {ok, []}
-    end.
 
 check_conf_sys(Env) ->
     check_conf_sys(Env, [], [], []).
@@ -1884,6 +1955,40 @@ load_file(File) ->
 	    {error, {none, open_file, "configuration file not found"}}
     end.
 
+load_file_descriptor(FileDescriptorId) ->
+    %% We do not want to read from the same file descriptor again if
+    %% init:restart is called so we use the old config obtained from
+    %% the file descriptor
+    case init:get_configfd(FileDescriptorId) of
+        none ->
+            WarningIntervalMs = 20000, % 20 seconds
+            MaxConfSizeBytes = 1024*1024*128, % 134 MB
+            case read_fd_until_end_and_close(FileDescriptorId,
+                                             WarningIntervalMs,
+                                             MaxConfSizeBytes) of
+                {ok, Bin} ->
+                    %% Make sure that there is some whitespace at the end of the string
+                    %% (so that reading a file with no NL following the "." will work).
+                    case file_binary_to_list(Bin) of
+                        {ok, String} ->
+                            case scan_file(String ++ " ") of
+                                {ok, Config} ->
+                                    init:set_configfd(FileDescriptorId, Config),
+                                    {ok, Config};
+                                Error ->
+                                    Error
+                            end;
+                        error ->
+                            {error, {none, scan_file, "bad encoding"}}
+                    end;
+                {error, Reason} ->
+                    {error, {none,
+                             read_from_file_descriptor,
+                             io_lib:format("Could not read from file descriptor: ~s", [Reason])}}
+            end;
+        Config -> {ok, Config}
+    end.
+
 scan_file(Str) ->
     case erl_scan:tokens([], Str, 1) of
 	{done, {ok, Tokens, _}, Left} ->
@@ -1908,6 +2013,61 @@ scan_file(Str) ->
 	    {error, {none, load_file, "no ending <dot> found"}}
     end.
 
+read_fd_until_end_and_close(FileDescriptorId,
+                            TimeBetweenWarningsMilliseconds,
+                            MaxSizeBytes) ->
+    ReadLimit = 1024,
+    Read =
+        fun Read(Ref, _Fd, _Data, DataSize) when DataSize > MaxSizeBytes ->
+                Reason = io_lib:format("Max size ~w bytes exceeded",
+                                       [MaxSizeBytes]),
+                {Ref, error, Reason};
+            Read(Ref, Fd, Data, DataSize) ->
+                case file:read(Fd, ReadLimit) of
+                    eof ->
+                        {Ref, ok, erlang:iolist_to_binary(Data)};
+                    {ok, Bin} ->
+                        Read(Ref, Fd, [Data, Bin], DataSize + byte_size(Bin));
+                    {error, Reason} ->
+                        {Ref, error, Reason}
+                end
+        end,
+    Receiver = self(),
+    Ref = make_ref(),
+    Reader =
+        spawn(
+          fun() ->
+                  case prim_file:file_desc_to_ref(FileDescriptorId, [read]) of
+                      {ok, Fd} ->
+                          Res = Read(Ref, Fd, [], 0),
+                          prim_file:close(Fd),
+                          Receiver ! Res;
+                      {error, _} ->
+                          Receiver ! {Ref,
+                                      error,
+                                      "Invalid file descriptor"}
+                  end
+          end),
+    Reader ! {reader_ref, Ref},
+    (fun GetResult() ->
+            receive
+                {Ref, ok, Data} ->
+                    {ok, erlang:iolist_to_binary(Data)};
+                {Ref, error, Reason} ->
+                    {error, Reason}
+            after TimeBetweenWarningsMilliseconds ->
+                    Msg =
+                        io_lib:format(
+                          "Slow -configfd file descriptor ~p. "
+                          "The system will continue to read from "
+                          "the file descriptor and will be blocked "
+                          "until end of file is received.",
+                          [FileDescriptorId]),
+                    ?LOG_WARNING(Msg),
+                    GetResult()
+            end
+    end)().
+
 only_ws([C|Cs]) when C =< $\s -> only_ws(Cs);
 only_ws([$%|Cs]) -> only_ws(strip_comment(Cs));   % handle comment
 only_ws([_|_]) -> false;
@@ -1930,9 +2090,12 @@ info_started(Name, Node) ->
                 report=>[{application, Name},
                          {started_at, Node}]},
               #{domain=>[otp,sasl],
-                report_cb=>fun logger:format_otp_report/1,
+                report_cb=>fun application_controller:format_log/2,
                 logger_formatter=>#{title=>"PROGRESS REPORT"},
-                error_logger=>#{tag=>info_report,type=>progress}}).
+                error_logger=>#{tag=>info_report,
+                                type=>progress,
+                                report_cb=>
+                                    fun application_controller:format_log/1}}).
 
 info_exited(Name, Reason, Type) ->
     ?LOG_NOTICE(#{label=>{application_controller,exit},
@@ -1940,8 +2103,140 @@ info_exited(Name, Reason, Type) ->
                            {exited, Reason},
                            {type, Type}]},
                 #{domain=>[otp],
-                  report_cb=>fun logger:format_otp_report/1,
-                  error_logger=>#{tag=>info_report,type=>std_info}}).
+                  report_cb=>fun application_controller:format_log/2,
+                error_logger=>#{tag=>info_report,
+                                type=>std_info,
+                                report_cb=>
+                                    fun application_controller:format_log/1}}).
+
+%% format_log/1 is the report callback used by Logger handler
+%% error_logger only. It is kept for backwards compatibility with
+%% legacy error_logger event handlers. This function must always
+%% return {Format,Args} compatible with the arguments in this module's
+%% calls to error_logger prior to OTP-21.0.
+format_log(LogReport) ->
+    Depth = error_logger:get_format_depth(),
+    FormatOpts = #{chars_limit => unlimited,
+                   depth => Depth,
+                   single_line => false,
+                   encoding => utf8},
+    format_log_multi(limit_report(LogReport, Depth), FormatOpts).
+
+limit_report(LogReport, unlimited) ->
+    LogReport;
+limit_report(#{label:={application_controller,progress},
+               report:=[{application,_}=Application,
+                        {started_at,Node}]}=LogReport,
+             Depth) ->
+    LogReport#{report=>[Application,
+                        {started_at,io_lib:limit_term(Node, Depth)}]};
+limit_report(#{label:={application_controller,exit},
+               report:=[{application,_}=Application,
+                        {exited,Reason},{type,Type}]}=LogReport,
+             Depth) ->
+    LogReport#{report=>[Application,
+                        {exited,io_lib:limit_term(Reason, Depth)},
+                        {type,io_lib:limit_term(Type, Depth)}]}.
+
+%% format_log/2 is the report callback for any Logger handler, except
+%% error_logger.
+format_log(Report, FormatOpts0) ->
+    Default = #{chars_limit => unlimited,
+                depth => unlimited,
+                single_line => false,
+                encoding => utf8},
+    FormatOpts = maps:merge(Default, FormatOpts0),
+    IoOpts =
+        case FormatOpts of
+            #{chars_limit:=unlimited} ->
+                [];
+            #{chars_limit:=Limit} ->
+                [{chars_limit,Limit}]
+        end,
+    {Format,Args} = format_log_single(Report, FormatOpts),
+    io_lib:format(Format, Args, IoOpts).
+
+format_log_single(#{label:={application_controller,progress},
+                    report:=[{application,Name},{started_at,Node}]},
+                  #{single_line:=true,depth:=Depth}=FormatOpts) ->
+    P = p(FormatOpts),
+    Format = "Application: "++P++". Started at: "++P++".",
+    Args =
+        case Depth of
+            unlimited ->
+                [Name,Node];
+            _ ->
+                [Name,Depth,Node,Depth]
+        end,
+    {Format,Args};
+format_log_single(#{label:={application_controller,exit},
+                    report:=[{application,Name},
+                             {exited,Reason},
+                             {type,Type}]},
+                  #{single_line:=true,depth:=Depth}=FormatOpts) ->
+    P = p(FormatOpts),
+    Format = lists:append(["Application: ",P,". Exited: ",P,
+                            ". Type: ",P,"."]),
+    Args =
+        case Depth of
+            unlimited ->
+                [Name,Reason,Type];
+            _ ->
+                [Name,Depth,Reason,Depth,Type,Depth]
+        end,
+    {Format,Args};
+format_log_single(Report,FormatOpts) ->
+    format_log_multi(Report,FormatOpts).
+
+format_log_multi(#{label:={application_controller,progress},
+                   report:=[{application,Name},
+                            {started_at,Node}]},
+                 #{depth:=Depth}=FormatOpts) ->
+    P = p(FormatOpts),
+    Format =
+        lists:append(
+          ["    application: ",P,"~n",
+           "    started_at: ",P,"~n"]),
+    Args =
+        case Depth of
+            unlimited ->
+                [Name,Node];
+            _ ->
+                [Name,Depth,Node,Depth]
+        end,
+    {Format,Args};
+format_log_multi(#{label:={application_controller,exit},
+                   report:=[{application,Name},
+                            {exited,Reason},
+                            {type,Type}]},
+                 #{depth:=Depth}=FormatOpts) ->
+    P = p(FormatOpts),
+    Format =
+        lists:append(
+          ["    application: ",P,"~n",
+           "    exited: ",P,"~n",
+           "    type: ",P,"~n"]),
+    Args =
+        case Depth of
+            unlimited ->
+                [Name,Reason,Type];
+            _ ->
+                [Name,Depth,Reason,Depth,Type,Depth]
+        end,
+    {Format,Args}.
+
+p(#{single_line:=Single,depth:=Depth,encoding:=Enc}) ->
+    "~"++single(Single)++mod(Enc)++p(Depth);
+p(unlimited) ->
+    "p";
+p(_Depth) ->
+    "P".
+
+single(true) -> "0";
+single(false) -> "".
+
+mod(latin1) -> "";
+mod(_) -> "t".
 
 %%-----------------------------------------------------------------
 %% Reply to all processes waiting this application to be started.  
